library(Interpol)
library(seqinr)

### calculates euclidean distance
euclidean <- function(a, b) sqrt(sum((a - b)^2))

### get p value for score distribution
getp <- function(score, distr){
  non.NA = length(which(is.na(distr)==FALSE))
  if(non.NA>0){
    tryCatch(
      {
        t = t.test(distr, mu=score)
        return(t$p.value)
      },
      error = function(cond){
        print("ERROR")
        return(NA)
      })
  }else{return(NA)}
}

### create database
createDatabase <- function(source){
  samples = length(source)
  sequences = rep(NA, samples)
  identifier = rep(NA, samples)
  lengths = rep(NA, samples)
  
  for(i in 1:samples){
    print(i)
    #print(no.samples)
    sequences[i] = source[[i]][1]
    identifier[i] = attr(source[[i]], "name")
    lengths[i] = length(unlist(strsplit(source[[i]][1],split="")))
  }
  
  database = data.frame(identifier, sequences, lengths)
  duplikate = which(duplicated(database[,2]))
  database = database[-duplikate,]
  return(database)
}


### Encode database
encodeDatabase <- function(datenbank, length_factor){
  l = length(datenbank[,2])
  encoded.database = matrix(NA, nrow=l, ncol=length_factor)
  for(i in 1:length(datenbank[,1])){
    a = Sys.time()
    s_enc = as.numeric(Interpol(AAdescriptor(datenbank[i,2], normalize = 2), length_factor))
    encoded.database[i,] = s_enc
    b = Sys.time()
    if(i%%1000==0){
      b = Sys.time()
      print(i)
      print(paste("Estimated time: ", (as.numeric(b-a))*(l-i)/60, " minutes"))
    }
    #print(i)
    #print(paste("Estimated time: ", (as.numeric(b-a))*(l-i)/60, " minutes"))
  }
  datenbank = data.frame(datenbank, encoded.database)
  return(datenbank)
}

### save database
saveDatabase <- function(file, database){
  save(file=file, list=c("database"))
  print("Database saved.")
}

# Search database
Interpol.search <- function(query, q_ol, datenbank, percentage){
  pun = 0
  length_factor = length(4:length(datenbank[1,]))
  q_enc = Interpol(AAdescriptor(query, normalize = 2), length_factor)
  best = c()
  #all_scores = c()   ### hier direkt alloziieren, damit schneller
  all_scores = rep(NA, nrow=length(datenbank[,1]))
  #######
  distance = 1000000000000
  ende = length_factor+3
  for(i in 1:length(datenbank[,1])){
    #print(i)
    #s_enc = Interpol(AAdescriptor(datenbank[i,2], normalize = 2), length_factor)
    #d = euclidean(q_enc, as.numeric(datenbank[i,4:ende])) + pun*(abs(q_ol - datenbank[i,3])) 
    if(datenbank[i,3]>(q_ol*(1+percentage))){
      #all_scores = c(all_scores, NA)
      next
    }
    if(datenbank[i,3]<(q_ol*(1-percentage))){
      #all_scores = c(all_scores, NA)
      next
    }
    d = euclidean(q_enc, as.numeric(datenbank[i,4:ende])) + (pun*(abs(q_ol - datenbank[i,3])))
    #d = pun*(abs(q_ol - datenbank[i,3])) 
    #### auskommentieren die nächste Zeile, nur für debugging
    all_scores[i] = d
    #######
    if(d<distance){
      distance = d
      best = i
    }
  }
  #print(paste("Best hit", best))
  #print(paste("Best score:", distance))
  return(list(best, distance, all_scores))  
}


##### get database hit
# query: query sequence as string
# database: database
# percentage: the maximum percentage the length is allowed to differ between sequence and target
getSequence <- function(query, database, percentage){
  q_ol = length(unlist(strsplit(query,split="")))
  subj = Interpol.search(query, q_ol, database, percentage)
  hit = data.frame(database[unlist(subj[[1]]),1:3], 
                   unlist(subj[[2]]), mean(unlist(subj[[3]]), na.rm = TRUE), 
                   getp(unlist(subj[[2]]), unlist(subj[[3]])))
  names(hit) = c("identifier", "sequence", "length", "score", "avgScore", "p")
  return(hit)
}

